<?php

namespace SGalinski\SgMail\Service;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use DateTime;
use SGalinski\SgMail\Domain\Model\Mail;
use SGalinski\SgMail\Domain\Model\Template;
use SGalinski\SgMail\Domain\Repository\LayoutRepository;
use SGalinski\SgMail\Domain\Repository\MailRepository;
use SGalinski\SgMail\Domain\Repository\TemplateRepository;
use Swift_Attachment;
use Swift_OutputByteStream;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Mail\MailMessage;
use TYPO3\CMS\Core\Resource\File;
use TYPO3\CMS\Core\Resource\ResourceFactory;
use TYPO3\CMS\Core\Routing\SiteMatcher;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\VersionNumberUtility;
use TYPO3\CMS\Extbase\Domain\Model\FileReference;
use TYPO3\CMS\Extbase\Object\ObjectManager;
use TYPO3\CMS\Extbase\Persistence\Generic\PersistenceManager;
use TYPO3\CMS\Extbase\Utility\LocalizationUtility;
use TYPO3\CMS\Fluid\View\StandaloneView;
use TYPO3\CMS\Frontend\Controller\TypoScriptFrontendController;

/**
 * MailTemplateService
 */
class MailTemplateService {
	const MARKER_TYPE_STRING = 'String';
	const MARKER_TYPE_ARRAY = 'Array';
	const MARKER_TYPE_OBJECT = 'Object';
	const MARKER_TYPE_FILE = 'File';
	const DEFAULT_LANGUAGE = 'default';
	const DEFAULT_TEMPLATE_PATH = 'Resources/Private/Templates/SgMail/';
	const CACHE_NAME = 'sg_mail_registerArrayCache';
	const CACHE_LIFETIME_IN_SECONDS = 86400;
	const REGISTER_FILE = 'Register.php';
	const CONFIG_PATH = 'Configuration/MailTemplates';

	/**
	 * @var string $toAddresses
	 */
	private $toAddresses = '';

	/**
	 * @var string $fromAddress
	 */
	private $fromAddress = '';

	/**
	 * @var string $ccAddresses
	 */
	private $ccAddresses;

	/**
	 * @var string $replyToAddress
	 */
	private $replyToAddress = '';

	/**
	 * @var string $language
	 */
	private $language = 'default';

	/**
	 * @var boolean $ignoreMailQueue
	 */
	private $ignoreMailQueue = FALSE;

	/**
	 * @var \TYPO3\CMS\Core\Mail\MailMessage $mailMessage
	 */
	private $mailMessage;

	/**
	 * @var string $templateName
	 */
	private $templateName;

	/**
	 * @var string $subject
	 */
	private $subject;

	/**
	 * @var string $overwrittenEmailBody
	 */
	private $overwrittenEmailBody = '';

	/**
	 * @var string $extensionKey
	 */
	private $extensionKey;

	/**
	 * @var array $markers
	 */
	private $markers;

	/**
	 * @var array $markerLabels
	 */
	private $markerLabels;

	/**
	 * @var string $bccAddresses
	 */
	private $bccAddresses;

	/**
	 * @var int
	 */
	private $priority = Mail::PRIORITY_LOWEST;

	/**
	 * @var int
	 */
	private $pid;

	/**
	 * @var string
	 */
	private $fromName = '';

	/**
	 * @var \SGalinski\SgMail\Domain\Repository\TemplateRepository
	 */
	protected $templateRepository;

	/**
	 * @var \SGalinski\SgMail\Domain\Repository\LayoutRepository
	 */
	protected $layoutRepository;

	/**
	 * @var \TYPO3\CMS\Extbase\Persistence\Generic\PersistenceManager
	 */
	protected $persistenceManager;

	/**
	 * @var \TYPO3\CMS\Extbase\Object\ObjectManager
	 */
	protected $objectManager;

	/**
	 * @var TYPO3\CMS\Core\Resource\ResourceFactory
	 */
	protected $resourceFactory;

	/**
	 * MailTemplateService constructor.
	 *
	 * @param string $templateName
	 * @param string $extensionKey
	 * @param array $markers
	 * @param array $markerLabels
	 * @throws \InvalidArgumentException
	 */
	public function __construct($templateName = '', $extensionKey = '', $markers = [], $markerLabels = []) {
		$this->templateName = $templateName;
		$this->extensionKey = $extensionKey;
		$this->markers = $markers;
		$this->markerLabels = $markerLabels;

		/** @var ObjectManager objectManager */
		$this->objectManager = GeneralUtility::makeInstance(ObjectManager::class);
		$this->mailMessage = $this->objectManager->get(MailMessage::class);
		$typoScriptSettingsService = $this->objectManager->get(TypoScriptSettingsService::class);
		$tsSettings = $typoScriptSettingsService->getSettings(0, 'tx_sgmail');
		$this->templateRepository = $this->objectManager->get(TemplateRepository::class);
		$this->layoutRepository = $this->objectManager->get(LayoutRepository::class);
		$this->persistenceManager = $this->objectManager->get(PersistenceManager::class);
		$this->resourceFactory = $this->objectManager->get(ResourceFactory::class);

		// use defaultMailFromAddress if it is provided in LocalConfiguration.php; use the sg_mail TS setting as fallback
		if (\filter_var($GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromAddress'], FILTER_VALIDATE_EMAIL)) {
			$this->fromAddress = $GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromAddress'];
		} else {
			$this->fromAddress = $tsSettings['mail']['default']['from'];

			if (!\filter_var($tsSettings['mail']['default']['from'], FILTER_VALIDATE_EMAIL)) {
				$this->fromAddress = 'noreply@example.org';
			} else {
				$this->fromAddress = $tsSettings['mail']['default']['from'];
			}
		}

		if ($GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromName']) {
			$this->fromName = $GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromName'];
		}

		$this->mailMessage->setFrom($this->fromAddress, $this->fromName);
		$this->bccAddresses = GeneralUtility::trimExplode(',', $tsSettings['mail']['default']['bcc']);
		$this->ccAddresses = GeneralUtility::trimExplode(',', $tsSettings['mail']['default']['cc']);

		foreach ($this->bccAddresses as $index => $email) {
			if (!\filter_var($email, FILTER_VALIDATE_EMAIL)) {
				unset($this->bccAddresses[$index]);
			}
		}

		foreach ($this->ccAddresses as $index => $email) {
			if (!filter_var($email, FILTER_VALIDATE_EMAIL)) {
				unset($this->ccAddresses[$index]);
			}
		}

		if (\count($this->bccAddresses) > 0) {
			$this->mailMessage->setBcc($this->bccAddresses);
		}

		if (\count($this->ccAddresses) > 0) {
			$this->mailMessage->setCc($this->ccAddresses);
		}
	}

	/**
	 * @param string $fromName
	 */
	public function setFromName($fromName) {
		$this->fromName = $fromName;
	}

	/**
	 * Provides translation for the marker data type
	 *
	 * @param string $markerType
	 */
	public static function getReadableMarkerType($markerType) {
		switch ($markerType) {
			case self::MARKER_TYPE_STRING :
				LocalizationUtility::translate('backend.marker.type.string', 'sg_mail');
				break;
			case self::MARKER_TYPE_ARRAY :
				LocalizationUtility::translate('backend.marker.type.array', 'sg_mail');
				break;
			case self::MARKER_TYPE_OBJECT :
				LocalizationUtility::translate('backend.marker.type.object', 'sg_mail');
				break;
			case self::MARKER_TYPE_FILE:
				LocalizationUtility::translate('backend.marker.type.file', 'sg_mail');
				break;
			default:
				LocalizationUtility::translate('backend.marker.type.mixed', 'sg_mail');
		}
	}

	/**
	 * @param string $toAddresses
	 * @return MailTemplateService
	 */
	public function setToAddresses($toAddresses): MailTemplateService {
		$normalizedToAddresses = trim(preg_replace('~\x{00a0}~iu', ' ', $toAddresses));
		$this->toAddresses = $normalizedToAddresses;

		$addressesArray = GeneralUtility::trimExplode(',', $normalizedToAddresses, TRUE);
		if (\count($addressesArray) > 1) {
			$normalizedToAddresses = $addressesArray;
		}
		$this->mailMessage->setTo($normalizedToAddresses);
		return $this;
	}

	/**
	 * @param string $fromAddress
	 * @param string $fromName
	 * @return MailTemplateService
	 */
	public function setFromAddress($fromAddress, $fromName = ''): MailTemplateService {
		if ($fromAddress) {
			$this->fromAddress = $fromAddress;
			$this->mailMessage->setFrom($fromAddress, $fromName);
		}

		return $this;
	}

	/**
	 * @param string $ccAddresses
	 * @return MailTemplateService
	 */
	public function setCcAddresses($ccAddresses): MailTemplateService {
		if ($ccAddresses) {
			$this->ccAddresses = $ccAddresses;
			$this->mailMessage->setCc(GeneralUtility::trimExplode(',', $this->ccAddresses));
		}

		return $this;
	}

	/**
	 * @param string $replyToAddress
	 * @return MailTemplateService
	 */
	public function setReplyToAddress($replyToAddress): MailTemplateService {
		if ($replyToAddress) {
			$this->replyToAddress = $replyToAddress;
			$this->mailMessage->setReplyTo($replyToAddress);
		}

		return $this;
	}

	/**
	 * @param string $language
	 * @return MailTemplateService
	 */
	public function setLanguage($language): MailTemplateService {
		$this->language = $language;
		return $this;
	}

	/**
	 * @param boolean $ignoreMailQueue
	 * @return MailTemplateService
	 */
	public function setIgnoreMailQueue($ignoreMailQueue): MailTemplateService {
		$this->ignoreMailQueue = $ignoreMailQueue;
		return $this;
	}

	/**
	 * @param string $templateName
	 * @return MailTemplateService
	 */
	public function setTemplateName($templateName): MailTemplateService {
		$this->templateName = $templateName;
		return $this;
	}

	/**
	 * @param string $extensionKey
	 * @return MailTemplateService
	 */
	public function setExtensionKey($extensionKey): MailTemplateService {
		$this->extensionKey = $extensionKey;
		return $this;
	}

	/**
	 * @param array $markers
	 * @return MailTemplateService
	 */
	public function addMarkers(array $markers): MailTemplateService {
		$this->setMarkers(\array_merge($this->markers, $markers));
		return $this;
	}

	/**
	 * @param array $markers
	 * @return MailTemplateService
	 */
	public function setMarkers(array $markers): MailTemplateService {
		$this->markers = $markers;
		foreach ($markers as $key => $currentMarker) {
			if (!\is_array($currentMarker) || !isset($currentMarker['markerLabel'])) {
				continue;
			}
			$this->markerLabels[$key] = $currentMarker['markerLabel'];
		}

		return $this;
	}

	/**
	 * @param string $bccAddresses
	 * @return MailTemplateService
	 */
	public function setBccAddresses($bccAddresses): MailTemplateService {
		if ($bccAddresses) {
			$this->bccAddresses = $bccAddresses;
			$this->mailMessage->setBcc(GeneralUtility::trimExplode(',', $this->bccAddresses));
		}

		return $this;
	}

	/**
	 * @param int $priority
	 * @return MailTemplateService
	 */
	public function setPriority($priority): MailTemplateService {
		$this->priority = $priority;
		return $this;
	}

	/**
	 * @param Swift_OutputByteStream $data
	 * @param string $filename
	 * @param string $contentType
	 * @return MailTemplateService
	 */
	public function addAttachment($data, $filename, $contentType): MailTemplateService {
		$attachment = Swift_Attachment::newInstance()
			->setFilename($filename)
			->setContentType($contentType)
			->setBody($data);
		$this->mailMessage->attach($attachment);
		return $this;
	}

	/**
	 * Adds a file resource as attachment
	 *
	 * @param FileReference $fileReference
	 * @param File $file
	 *
	 * @return MailTemplateService
	 * @throws \TYPO3\CMS\Core\Resource\Exception\ResourceDoesNotExistException
	 */
	public function addFileResourceAttachment(
		FileReference $fileReference = NULL, File $file = NULL
	): MailTemplateService {
		if (!$file) {
			if (!$fileReference) {
				return $this;
			}

			$originalResource = $fileReference->getOriginalResource();
			if (!$originalResource) {
				return $this;
			}

			$file = $originalResource->getOriginalFile();
			if (!$file) {
				return $this;
			}
		}

		$coreFileReferenceMailFile = $this->resourceFactory->createFileReferenceObject(
			[
				'uid_local' => $file->getUid(),
				'table_local' => 'sys_file',
				'uid' => uniqid('NEW_MAIL', TRUE)
			]
		);

		$newFileReference = GeneralUtility::makeInstance(FileReference::class);
		$newFileReference->setOriginalResource($coreFileReferenceMailFile);

		$this->markers[] = $newFileReference;
		return $this;
	}

	/**
	 * @return MailMessage
	 */
	public function getMailMessage(): MailMessage {
		return $this->mailMessage;
	}

	/**
	 * set the page id from which this was called
	 *
	 * @param int $pid
	 * @return MailTemplateService
	 */
	public function setPid($pid): MailTemplateService {
		$this->pid = (int) $pid;
		return $this;
	}

	/**
	 * Checks if a template is blacklisted for a given siteroot id
	 *
	 * @param string $extensionKey
	 * @param string $templateName
	 * @param int $siteRootId
	 * @return boolean
	 * @throws \InvalidArgumentException
	 * @throws \BadFunctionCallException
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 */
	public static function isTemplateBlacklisted($extensionKey, $templateName, $siteRootId): bool {
		$nonBlacklistedTemplates = BackendService::getNonBlacklistedTemplates($siteRootId);
		if ($nonBlacklistedTemplates[$extensionKey]) {
			return $nonBlacklistedTemplates[$extensionKey][$templateName] ? FALSE : TRUE;
		}

		return TRUE;
	}

	/**
	 * @return string
	 */
	public function getSubject(): string {
		return $this->subject;
	}

	/**
	 * @param string $subject
	 */
	public function setSubject(string $subject) {
		$this->subject = $subject;
	}

	/**
	 * @return string
	 */
	public function getOverwrittenEmailBody(): string {
		return $this->overwrittenEmailBody;
	}

	/**
	 * @param string $overwrittenEmailBody
	 */
	public function setOverwrittenEmailBody(string $overwrittenEmailBody) {
		$this->overwrittenEmailBody = $overwrittenEmailBody;
	}

	/**
	 * Return default markers for sg_mail
	 *
	 * @param string $translationKey
	 * @param array $marker
	 * @param string $extensionKey
	 * @return array
	 */
	public static function getDefaultTemplateMarker($translationKey, array $marker, $extensionKey = 'sg_mail'): array {
		$languagePath = 'LLL:EXT:' . $extensionKey . '/Resources/Private/Language/locallang.xlf:' . $translationKey;
		// Need the key for translations
		if (\trim($extensionKey) === '') {
			return [];
		}

		$generatedMarker = [];
		foreach ($marker as $markerName) {
			$generatedMarker[] = [
				'marker' => $markerName,
				'value' => $languagePath . '.example.' . $markerName,
				'description' => $languagePath . '.description.' . $markerName,
				'backend_translation_key' => $translationKey . '.example.' . $markerName,
				'extension_key' => $extensionKey
			];
		}

		return $generatedMarker;
	}

	/**
	 * Send the Email
	 *
	 * @param bool $isPreview
	 * @return bool email was sent or added to mail queue successfully?
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 * @throws \InvalidArgumentException
	 * @throws \BadFunctionCallException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException
	 * @throws \Exception
	 */
	public function sendEmail($isPreview = FALSE): bool {
		$success = FALSE;
		if (TYPO3_MODE === 'FE') {
			/** @var TypoScriptFrontendController $typoscriptFrontendController */
			$typoscriptFrontendController = $GLOBALS['TSFE'];
			$pageUid = (int) $typoscriptFrontendController->id;
		} else {
			$pageUid = (int) GeneralUtility::_GP('id');
		}

		if ($this->pid) {
			$pageUid = $this->pid;
		}

		if ($pageUid === 0) {
			$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable(
				'pages'
			);
			$rootPageRows = $queryBuilder->select('*')
				->from('pages')
				->where(
					$queryBuilder->expr()->eq(
						'is_siteroot', 1
					)
				)
				->andWhere(
					$queryBuilder->expr()->eq(
						'hidden', 0
					)
				)
				->execute()->fetchAll();

			if ($rootPageRows && \count($rootPageRows)) {
				$pageUid = (int) $rootPageRows[0]['uid'];
			}
		}
		$siteRootId = BackendService::getSiteRoot($pageUid);

		$isTemplateBlacklisted = self::isTemplateBlacklisted($this->extensionKey, $this->templateName, $siteRootId);
		if ($isTemplateBlacklisted) {
			// @TODO throw error or log ?
			return FALSE;
		}

		/** @var Template $template */
		$template = $this->templateRepository->findOneByTemplate(
			$this->extensionKey, $this->templateName, $this->language, $siteRootId
		);

		if ($template === NULL) {
			$template = $this->templateRepository->findOneByTemplate(
				$this->extensionKey, $this->templateName, 'default', $siteRootId
			);
		}

		// if there is a template, prefer those values
		if ($template) {
			$this->loadTemplateValues($template);
		}

		// get default template content from register array
		$registerService = GeneralUtility::makeInstance(RegisterService::class);
		$defaultTemplateContent =
			$registerService->getRegisterArray()[$this->extensionKey][$this->templateName]['templateContent'];

		// If there is no template for this language, use the default template
		if ($template === NULL) {
			if ($defaultTemplateContent === NULL) {
				$templatePath =
					$registerService->getRegisterArray()[$this->extensionKey][$this->templateName]['templatePath'];

				// only standard template file is considered since version 4.1
				$defaultTemplateFile = $templatePath . 'template.html';
				if (\file_exists($defaultTemplateFile)) {
					$defaultTemplateContent = \file_get_contents($defaultTemplateFile);
				} else {
					// use configured default html template
					/** @var TypoScriptSettingsService $typoScriptSettingsService */
					$typoScriptSettingsService = $this->objectManager->get(TypoScriptSettingsService::class);
					$tsSettings = $typoScriptSettingsService->getSettings(0, 'tx_sgmail');
					$defaultTemplateFile = GeneralUtility::getFileAbsFileName(
						$tsSettings['mail']['defaultHtmlTemplate']
					);

					if (\file_exists($defaultTemplateFile)) {
						$defaultTemplateContent = \file_get_contents($defaultTemplateFile);
					} else {
						return FALSE;
					}
				}
			}
		} elseif (\filter_var($template->getToAddress(), FILTER_VALIDATE_EMAIL)) {
			$this->setToAddresses(\trim($template->getToAddress()));
		}

		if ($isPreview) {
			$previewMarker = [];

			/** @var array $markerArray */
			$markerArray = $registerService->getRegisterArray()[$this->extensionKey][$this->templateName]['marker'];
			foreach ($markerArray as $marker) {
				$markerPath = GeneralUtility::trimExplode('.', $marker['marker']);
				$temporaryMarkerArray = [];
				foreach (\array_reverse($markerPath) as $index => $markerPathSegment) {
					if ($index === 0) {
						if ($marker['markerLabel']) {
							$markerPathSegment = $marker['markerLabel'];
						}

						if ($marker['backend_translation_key']) {
							$temporaryMarkerArray[$markerPathSegment] = LocalizationUtility::translate(
								$marker['backend_translation_key'], $marker['extension_key']
							);
						} else {
							$temporaryMarkerArray[$markerPathSegment] = $marker['value'];
						}
					} else {
						$temporaryMarkerArray = [$markerPathSegment => $temporaryMarkerArray];
					}
				}
				/** @noinspection SlowArrayOperationsInLoopInspection */
				$previewMarker = \array_merge_recursive($previewMarker, $temporaryMarkerArray);
			}
			$this->setIgnoreMailQueue(TRUE);
			$this->setMarkers($previewMarker);
		}

		/** @var StandaloneView $emailView */
		$emailView = $this->objectManager->get(StandaloneView::class);
		$emailView->assignMultiple($this->markers);
		$emailView->assign('all_fields', $this->getAllMarker($this->markers));

		$overwrittenEmailBody = $this->getOverwrittenEmailBody();
		$overwrittenSubject = '';
		if ($this->subject !== '' && $this->subject !== NULL) {
			$overwrittenSubject = $this->subject;
		}

		if ($template !== NULL) {
			$emailView->setTemplateSource(
				\trim(empty($overwrittenSubject) ? $template->getSubject() : $overwrittenSubject)
			);
			$subject = $emailView->render();

			$emailView->setTemplateSource(
				$this->getTemplateSource(
					empty($overwrittenEmailBody) ? $template->getContent() : $overwrittenEmailBody,
					$template->getLayout(), $siteRootId
				)
			);
		} else {
			$subject = $registerService->getRegisterArray()[$this->extensionKey][$this->templateName]['subject'];
			if (\is_array($subject)) {
				$subject = \trim(
					$registerService->getRegisterArray()
					[$this->extensionKey][$this->templateName]['subject'][$this->language]
				);
			}

			$emailView->setTemplateSource(empty($overwrittenSubject) ? $subject : $overwrittenSubject);
			$subject = $emailView->render();

			$emailView->setTemplateSource(
				$this->getTemplateSource(
					empty($overwrittenEmailBody) ? $defaultTemplateContent : $overwrittenEmailBody,
					0, $siteRootId
				)
			);
		}

		$this->mailMessage->setSubject($subject);

		// insert <br> tags, but replace every instance of three or more successive breaks with just two.
		$emailBody = $emailView->render();
		$emailBody = \nl2br($emailBody);
		$emailBody = \preg_replace('/(<br[\s]?[\/]?>[\s]*){3,}/', '<br><br>', $emailBody);

		$mail = $this->addMailToMailQueue(
			$this->extensionKey, $this->templateName, $subject, $emailBody, $this->priority,
			0, 0, $this->language, $siteRootId
		);
		if ($this->ignoreMailQueue) {
			$success = $this->sendMailFromQueue($mail->getUid());
		}

		if ($isPreview) {
			$mailRepository = $this->objectManager->get(MailRepository::class);
			$mailRepository->remove($mail);
			$this->persistenceManager->persistAll();
		}

		return $success;
	}

	/**
	 * Combines the template content with the layout and returns the result
	 *
	 * @param string $content
	 * @param int $layoutUid
	 * @param int $siteRootId
	 * @return string
	 */
	private function getTemplateSource(string $content, int $layoutUid, int $siteRootId): string {
		if ($layoutUid === -1) {
			return $content;
		}

		$languageUid = 0;
		if ($this->language !== self::DEFAULT_LANGUAGE) {
			$languageUid = (int) array_search($this->language, $this->getAvailableLanguages(), TRUE);
		}

		$frontendSimulated = FALSE;
		if (!isset($GLOBALS['TSFE'])) {
			$frontendSimulated = TRUE;
			$GLOBALS['TSFE'] = new TypoScriptFrontendController(NULL, $siteRootId, 0);
		}
		/** @var array $layout */
		$layout = $this->layoutRepository->findByUidOrDefault($layoutUid, $siteRootId, $languageUid);
		if ($frontendSimulated) {
			unset($GLOBALS['TSFE']);
		}

		if ($layout === NULL) {
			return $content;
		}

		return str_replace('###CONTENT###', $content, $layout['content']);
	}

	/**
	 * Returns the list of avalilable translation languages
	 *
	 * @return array
	 */
	private function getAvailableLanguages(): array {
		$out = [0 => ''];
		if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '9.0.0', '<')) {
			$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable(
				'sys_language'
			);
			$rows = $queryBuilder->select('*')
				->from('sys_language')->execute()->fetchAll();

			foreach ($rows as $row) {
				$out[(int) $row['uid']] = $row['language_isocode'];
			}
		} else {
			try {
				$site = GeneralUtility::makeInstance(SiteMatcher::class)->matchByPageId(0);
			} catch (\Exception $exception) {
				return [0 => ''];
			}
			$availableLanguages = $site->getLanguages();
			$out = [];
			foreach ($availableLanguages as $language) {
				$languageId = $language->getLanguageId();
				if ($languageId < 0) {
					continue;
				}
				$out[$language->getLanguageId()] = strtolower($language->getTwoLetterIsoCode());
			}
		}
		return $out;
	}

	/**
	 * Adds a new mail to the mail queue.
	 *
	 * @param string $extensionKey
	 * @param string $templateName
	 * @param string $subject
	 * @param string $emailBody
	 * @param int $sendingTime
	 * @param int $priority
	 * @param int $lastSendingTime
	 * @param string $language
	 * @param int $pid
	 * @return Mail
	 * @throws \InvalidArgumentException
	 * @throws \BadFunctionCallException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException
	 * @throws \TYPO3\CMS\Core\Cache\Exception\NoSuchCacheException
	 */
	private function addMailToMailQueue(
		$extensionKey, $templateName, $subject, $emailBody, $priority, $sendingTime = 0,
		$lastSendingTime = 0, $language = self::DEFAULT_LANGUAGE, $pid = 0
	): Mail {
		$mail = $this->objectManager->get(Mail::class);
		$mail->setPid($pid);
		$mail->setExtensionKey($extensionKey);
		$mail->setTemplateName($templateName);
		$mail->setLanguage($language);
		$mail->setBlacklisted(self::isTemplateBlacklisted($extensionKey, $templateName, $pid));

		$mail->setFromAddress($this->fromAddress);
		$mail->setFromName($this->fromName);

		$mail->setToAddress($this->toAddresses);
		$mail->setMailSubject($subject);
		$mail->setMailBody($emailBody);
		$mail->setPriority($priority);
		$mail->setBccAddresses($this->bccAddresses);
		$mail->setCcAddresses($this->ccAddresses);
		$mail->setSendingTime($sendingTime);
		$mail->setLastSendingTime($lastSendingTime);
		$mail->setReplyTo($this->replyToAddress);
		foreach ($this->markers as $marker) {
			if ($marker instanceof FileReference) {
				$mail->addAttachment($marker);
			}
		}

		$mailRepository = $this->objectManager->get(MailRepository::class);
		$mailRepository->add($mail);
		$this->persistenceManager->persistAll();
		return $mail;
	}

	/**
	 * Send a Mail from the queue, identified by its id
	 *
	 * @param int $uid
	 * @return bool|NULL
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\UnknownObjectException
	 * @throws \Exception
	 */
	public function sendMailFromQueue($uid): bool {
		$mailRepository = $this->objectManager->get(MailRepository::class);
		/** @var Mail $mailToSend */
		$mailToSend = $mailRepository->findOneByUid($uid);
		if (!$mailToSend || $mailToSend->getBlacklisted()) {
			return FALSE;
		}

		$this->mailMessage->setBody($mailToSend->getMailBody(), 'text/html');
		$plaintextService = GeneralUtility::makeInstance(PlaintextService::class);
		$plaintextBody = $plaintextService->makePlain($mailToSend->getMailBody());
		$this->mailMessage->addPart($plaintextBody, 'text/plain');
		$toAddresses = \trim($mailToSend->getToAddress());
		$addressesArray = GeneralUtility::trimExplode(',', $toAddresses, TRUE);
		if (\count($addressesArray) > 1) {
			$toAddresses = $addressesArray;
		}
		$this->mailMessage->setTo($toAddresses);
		$this->mailMessage->setFrom($mailToSend->getFromAddress(), $mailToSend->getFromName());
		$this->mailMessage->setSubject($mailToSend->getMailSubject());

		if ($mailToSend->getBccAddresses()) {
			$this->mailMessage->setBcc(GeneralUtility::trimExplode(',', $mailToSend->getBccAddresses()));
		}

		if ($mailToSend->getCcAddresses()) {
			$this->mailMessage->setCc(GeneralUtility::trimExplode(',', $mailToSend->getCcAddresses()));
		}

		if ($mailToSend->getReplyTo()) {
			$this->mailMessage->setReplyTo($mailToSend->getReplyTo());
		}
		$attachments = $mailToSend->getAttachments();
		if ($attachments->count() > 0) {
			foreach ($attachments as $attachment) {
				/**
				 * @var FileReference $attachment
				 */
				$file = $attachment->getOriginalResource()->getOriginalFile();
				$this->mailMessage->attach(
					\Swift_Attachment::newInstance($file->getContents(), $file->getName(), $file->getMimeType())
				);
			}
		}
		$dateTime = new DateTime();
		if ((int) $mailToSend->getSendingTime() === 0) {
			$mailToSend->setSendingTime($dateTime->getTimestamp());
		}
		$mailToSend->setLastSendingTime($dateTime->getTimestamp());
		$success = $this->mailMessage->send();
		if ($success) {
			$mailRepository->update($mailToSend);
		} else {
			$this->mailMessage->getFailedRecipients();
		}
		return $success;
	}

	/**
	 * use all values from the given template
	 *
	 * @param Template $template
	 */
	private function loadTemplateValues($template) {
		$fromName = \trim($template->getFromName());
		if ($fromName === '') {
			$fromName = $this->fromName;
		}
		if ($fromName === '' && $GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromName']) {
			$fromName = $GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromName'];
		}

		$fromMail = $this->getValidFromMail(\trim($template->getFromMail()));
		$this->setFromAddress($fromMail, $fromName);
		$this->setCcAddresses($template->getCc());
		$this->setBccAddresses($template->getBcc());
		$this->setReplyToAddress($template->getReplyTo());
		$this->setFromName($fromName);
	}

	/**
	 * Sets the fromMail property of the mailTemplateService.
	 * Checks validity and uses all available fallbacks
	 *
	 * @param string $fromMail
	 * @return string
	 */
	private function getValidFromMail($fromMail): string {
		$fromMail = \trim($fromMail);
		if (!\filter_var($fromMail, FILTER_VALIDATE_EMAIL)) {
			$fromMail = $this->fromAddress;
		}
		if (!\filter_var($fromMail, FILTER_VALIDATE_EMAIL)) {
			$fromMail = $GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromAddress'];
			if (!\filter_var($GLOBALS['TYPO3_CONF_VARS']['MAIL']['defaultMailFromAddress'], FILTER_VALIDATE_EMAIL)) {
				$fromMail = 'noreply@example.com';
			}
		}

		return $fromMail;
	}

	/**
	 * Get a single variable containing a list of all markers
	 *
	 * @param array $markers
	 * @return string
	 */
	private function getAllMarker(array $markers): string {
		$allMarker = '';

		foreach ($markers as $key => $value) {
			if (\array_key_exists($key, $this->markerLabels) && $this->markerLabels[$key] !== NULL) {
				$key = $this->markerLabels[$key];
			}

			if (\is_string($value)) {
				$allMarker .= $key . ': ' . $value . PHP_EOL;
			} elseif (\is_array($value)) {
				foreach ($value as $innerKey => $innerValue) {
					$allMarker .= $key . '.' . $innerKey . ': ' . $innerValue . PHP_EOL;
				}
			} elseif (\is_bool($value)) {
				$valueAsString = $value ? 'true' : 'false';
				$allMarker .= $key . ': ' . $valueAsString . PHP_EOL;
			} elseif (\is_object($value)) {
				if (\method_exists($value, '__toString')) {
					$allMarker .= $key . ': ' . $value->__toString() . PHP_EOL;
				}
			}
		}

		return $allMarker;
	}
}
